import os
import pathlib
import subprocess

import necxxt.utils
import necxxt.utils.logging

__env__ = {
    "CXX": "docker run -ti -v {}:/necxxt --rm necxxt/clang".format(os.getcwd()),
    "CXXFLAGS": "-std=c++2a -fmodules-ts"
}


def runCompiler(cmds):
    proc = subprocess.Popen(cmds)
    proc.communicate()

    if proc.returncode != 0:
        exit(proc.returncode)


def checkIfModulesAreInstalled(modules):
    for m in modules:
        p = os.path.join("necxxt_modules", m)
        if not os.path.isdir(p):
            necxxt.utils.logging.error(
                "Module {} is not installed".format({m}))
            exit(1)


def precompileModules(order, build_modules):
    for dependency in order:
        p = os.path.join("necxxt_modules", dependency)
        for f in necxxt.utils.listFiles(p, filters=[".cxxm"], shorten=False):
            i = f
            d = os.path.dirname(f)
            f = os.path.basename(f)
            name, extension = os.path.splitext(f)
            o = os.path.join("build", d, name + ".pcm")
            pathlib.Path(os.path.dirname(o)).mkdir(parents=True)

            modules = []
            for _m in build_modules.get(dependency, []):
                for _f in necxxt.utils.listFiles(os.path.join("build", "necxxt_modules",
                                                              _m), filters=[".pcm"], shorten=False):
                    if _f not in modules:
                        modules.append(_f)

            cmds = __env__["CXX"].split(" ")
            cmds.extend(__env__["CXXFLAGS"].split(" "))
            cmds.append("--precompile")
            cmds.append(i)

            for m in modules:
                cmds.append("-fmodule-file={}".format(m))

            cmds.append("-o")
            cmds.append(o)

            necxxt.utils.logging.debug(" ".join(cmds))
            runCompiler(cmds)


def compileModules(order, build_modules):
    for f in necxxt.utils.listFiles(os.getcwd(), filters=[".pcm"]):
        i = f
        d = os.path.dirname(f)
        f = os.path.basename(f)
        name, extension = os.path.splitext(f)
        o = os.path.join(d, name + ".pcm.o")

        cmds = __env__["CXX"].split(" ")
        cmds.extend(__env__["CXXFLAGS"].split(" "))
        cmds.append("-c")
        cmds.append(i)
        cmds.append("-o")
        cmds.append(o)

        necxxt.utils.logging.debug(" ".join(cmds))
        runCompiler(cmds)


def compileSources(order, build_modules):
    modules = []
    for f in necxxt.utils.listFiles(os.getcwd(), filters=[".pcm"]):
        if f not in modules:
            modules.append(f)

    for f in necxxt.utils.listFiles(os.getcwd(), filters=[".cxx"]):
        i = f
        d = os.path.dirname(f)
        f = os.path.basename(f)
        name, extension = os.path.splitext(f)
        o = os.path.join("build", d, name + ".cxx.o")
        pathlib.Path(os.path.dirname(o)).mkdir(parents=True)

        cmds = __env__["CXX"].split(" ")
        cmds.extend(__env__["CXXFLAGS"].split(" "))
        cmds.append("-c")
        cmds.append(i)

        for m in modules:
            cmds.append("-fmodule-file={}".format(m))

        cmds.append("-o")
        cmds.append(o)

        necxxt.utils.logging.debug(" ".join(cmds))
        runCompiler(cmds)


def buildStatic(order, build_modules):
    cmds = __env__["CXX"].split(" ")
    cmds.extend(__env__["CXXFLAGS"].split(" "))

    module_paths = {}
    objects = []
    for o in order:
        for f in necxxt.utils.listFiles(os.getcwd(), filters=[".pcm.o"]):
            d = os.path.dirname(f)
            if o in d:
                if d not in module_paths:
                    module_paths[d] = []

                if f not in module_paths[d]:
                    module_paths[d].append(f)

    for f in necxxt.utils.listFiles(os.getcwd(), filters=[".cxx.o"]):
        if f not in objects:
            objects.append(f)

    for module_path in module_paths.keys():
        cmds.append("-fprebuilt-module-path={}".format(module_path))

        for o in module_paths.get(module_path, []):
            cmds.append(o)

    for o in objects:
        cmds.append(o)

    cmds.append("-o")

    module = necxxt.utils.readModule()
    cmds.append(os.path.join("build", module["name"]))
    cmds.append("-static")

    necxxt.utils.logging.debug(" ".join(cmds))
    runCompiler(cmds)



import io
import json
import os
import pathlib
import shutil
import urllib.request
import zipfile


class Repository:
    def __init__(self, name, version):
        self._name = name
        self._version = version

    def readLocalModule(self):
        module_file = os.path.join("necxxt_modules", self._name, "necxxt.json")
        with open(module_file, "r") as f:
            return json.load(f)

    def readModule(self):
        pass

    def install(self):
        pass

    def __repr__(self):
        return None

    def __str__(self):
        return self.__repr__()


class RepositoryDirectory(Repository):
    def readModule(self):
        module_file = os.path.join(self._version, "necxxt.json")
        with open(module_file, "r") as f:
            return json.load(f)

    def install(self):
        pathlib.Path("necxxt_modules").mkdir(parents=True, exist_ok=True)

        if os.path.isdir(os.path.join("necxxt_modules", self._name)):
            shutil.rmtree(os.path.join("necxxt_modules", self._name))

        shutil.copytree(self._version, os.path.join(
            "necxxt_modules", self._name))

    def __repr__(self):
        return self._version


class RepositoryGitHub(Repository):
    def readModule(self):
        url = "https://raw.githubusercontent.com/necxxt/{}/master/necxxt.json".format(
            self._name)
        response = urllib.request.urlopen(url)
        if response.getcode() == 200:
            data = response.read()
            encoding = response.info().get_content_charset("utf-8")
            return json.loads(data.decode(encoding))

    def install(self):
        url = str(self)
        response = urllib.request.urlopen(url)
        if response.getcode() == 200:
            data = response.read()
            z = zipfile.ZipFile(io.BytesIO(data))

            if os.path.isdir(os.path.join("necxxt_modules", self._name)):
                shutil.rmtree(os.path.join("necxxt_modules", self._name))

            if os.path.isdir(os.path.join("necxxt_modules", "{}-master".format(self._name))):
                shutil.rmtree(os.path.join("necxxt_modules",
                                           "{}-master".format(self._name)))

            z.extractall("necxxt_modules")

            shutil.move(os.path.join("necxxt_modules", "{}-master".format(self._name)),
                        os.path.join("necxxt_modules", self._name))

    def __repr__(self):
        return "https://github.com/necxxt/{}/archive/master.zip".format(self._name)


def resolveRepository(name, version):
    if os.path.isdir(version):
        return RepositoryDirectory(name, version)
    elif isHostedOnGitHub(name):
        return RepositoryGitHub(name, version)

    return None


def isHostedOnGitHub(name):
    url = "https://github.com/necxxt/{}".format(name)
    response = urllib.request.urlopen(url)

    return response.getcode() == 200



import json

import necxxt.repositories
import necxxt.utils


def resolveDepency(name, version):
    return necxxt.repositories.resolveRepository(name, version)


def getResolvedDependencies(graph):
    resolved = []

    if graph.get("requires", False):
        dependencies = graph["dependencies"].keys()
        for dependency in dependencies:
            if "resolved" in dependencies[dependency]:
                resolved.append(dependency)

    return resolved


def areAllDependenciesResolved(graph):
    if graph.get("requires", False):
        dependencies = graph["dependencies"].keys()
        resolved_dependencies = 0
        for dependency in dependencies:
            if "resolved" in dependencies[dependency]:
                resolved_dependencies += 1

        return resolved_dependencies == len(dependencies)

    return False


def installDependencies():
    graph = {}

    module = necxxt.utils.readModule()

    graph["name"] = module["name"]
    graph["version"] = module["version"]
    graph["lockfileVersion"] = 1

    if not "dependencies" in module:
        return graph

    graph["requires"] = True
    graph["dependencies"] = {}

    dependencies = list(module["dependencies"].items())
    i = 0
    while i < len(dependencies):
        dependency, version = dependencies[i]
        if dependency not in graph["dependencies"]:
            graph["dependencies"][dependency] = {
                "version": version
            }

            resolved = resolveDepency(dependency, version)
            if resolved:
                graph["dependencies"][dependency]["resolved"] = str(resolved)
                resolved.install()

                m = resolved.readLocalModule()
                requires = m.get("dependencies")
                if requires:
                    graph["dependencies"][dependency]["requires"] = requires

                    for d, v in requires.items():
                        if not d in graph["dependencies"]:
                            dependencies.append((d, v,))

        i += 1

    with open("necxxt-lock.json", "w") as f:
        json.dump(graph, f, indent=4)

    return graph



import json
import os

import necxxt.utils.logging


def listCompilers():
    compilers = []

    for path in os.environ["PATH"].split(os.pathsep):
        compilers.extend(
            listFiles(path, filters=["clang++", "g++"], shorten=False))

    return compilers


def listFiles(d, filters=[], shorten=True):
    fs = []

    for path, subdirs, files in os.walk(d):
        for name in files:
            add = False
            if len(filters) == 0:
                add = True
            else:
                for f in filters:
                    if name.endswith(f):
                        add = True

            if add:
                p = os.path.join(path, name)
                if shorten:
                    fs.append(os.path.relpath(p, d))
                else:
                    fs.append(p)

    return fs


def readModule(dependency=None):
    if dependency:
        module_file = os.path.join(
            os.getcwd(), "necxxt_modules", dependency, "necxxt.json")
    else:
        module_file = os.path.join(os.getcwd(), "necxxt.json")

    if not os.path.isfile(module_file):
        necxxt.utils.logging.error(
            "ENOENT: no such file or directory, open '{}'".format(module_file))
        exit(1)

    module = {}
    with open(module_file, "r") as f:
        module = json.load(f)

    return module



import sys


class colors:
    END = "\033[0m"
    ERROR = "\033[1;31m"
    INFO = "\033[1;32m"
    WARNING = "\033[1;33m"
    DEBUG = "\033[1;34m"


def error(msg):
    sys.stderr.write(colors.ERROR + msg + colors.END + "\n")


def info(msg):
    print(colors.INFO + msg + colors.END)


def warn(msg):
    print(colors.WARNING + msg + colors.END)


def debug(msg):
    print(colors.DEBUG + msg + colors.END)



import os
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open(os.path.join("necxxt", "version.py"), "r") as f:
    exec(f.read())

setuptools.setup(
    name="necxxt",
    version=__version__,
    author="Marius Meisenzahl",
    author_email="mariusmeisenzahl@gmail.com",
    description="Tooling for next level C++ projects",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/necxxt/cli",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    scripts=["bin/necxxt"],
)



import copy
import json
import os
import pathlib
import shutil
import subprocess

from .build import checkIfModulesAreInstalled, precompileModules, compileModules, compileSources, buildStatic

import necxxt.dependencies
from necxxt.version import __executable__


class Command:
    def __init__(self):
        self._name = None

    def run(self):
        pass

    def __str__(self):
        return self._name


class HelpCommand(Command):
    def __init__(self):
        self._name = "help"
        self._executable = "necxxt"

    def run(self):
        print()
        print("Usage: {} <command>".format(__executable__))
        print()
        print("where <command> is one of:")
        print("    build, help, install")
        print()
        # print("Specify configs in the ini-formatted file:")
        # print("    {}".format(os.path.join(
        #     pathlib.Path.home(), ".{}rc".format(__executable__))))
        # print(
        #     "or on the command line via: {} < command > --key value".format(__executable__))
        # print("Config info can be viewed via: {} help config".format(__executable__))
        # print()
        print("{}@{} {}".format(__executable__, necxxt.__version__, __file__))


class BuildCommand(Command):
    def __init__(self):
        self._name = "build"

    def run(self):
        if os.path.isdir("build"):
            shutil.rmtree("build")

        lock_path = os.path.join(os.getcwd(), "necxxt-lock.json")
        if not os.path.isfile(lock_path):
            necxxt.utils.logging.error(
                "ENOENT: no such file or directory, open '{}'".format(lock_path))
            exit(1)

        with open(lock_path, "r") as f:
            lock = json.load(f)

        modules = {}
        for dependency in lock["dependencies"].keys():
            requires = lock["dependencies"][dependency].get("requires", {})
            modules[dependency] = list(requires.keys())

        build_modules = copy.deepcopy(modules)

        order = []
        i = 0
        while i < len(modules.keys()):
            keys = list(modules.keys())
            key = keys[i]
            dependencies = modules[key]

            if len(dependencies) == 0:
                order.append(key)
                del modules[key]
                for d in modules.keys():
                    if key in modules[d]:
                        modules[d].remove(key)

            i += 1

            length = len(modules.keys())
            if length > 0 and i >= length:
                i = 0

        checkIfModulesAreInstalled(order)
        precompileModules(order, build_modules)
        compileModules(order, build_modules)
        compileSources(order, build_modules)
        buildStatic(order, build_modules)


class InstallCommand(Command):
    def __init__(self):
        self._name = "install"

    def run(self):
        necxxt.dependencies.installDependencies()


class RunCommand(Command):
    def __init__(self):
        self._name = "run"

    def run(self):
        lock_path = os.path.join(os.getcwd(), "necxxt-lock.json")
        if not os.path.isfile(lock_path):
            necxxt.utils.logging.error(
                "ENOENT: no such file or directory, open '{}'".format(lock_path))
            exit(1)

        with open(lock_path, "r") as f:
            lock = json.load(f)

        exe = os.path.join(os.getcwd(), "build", lock.get("name", ""))
        if not os.path.isfile(exe):
            necxxt.utils.logging.error(
                "ENOENT: no such file or directory, open '{}'".format(exe))
            exit(1)

        subprocess.call([exe])
